"""Miscellaneous functions used in wepy that don't fit elsewhere.

Warnings
--------

Deprecation warning!

Try not to rely on these functions as part of your stable code. These
were included here in wepy because other libraries do not provide this
functionality, although they probably should. That is these functions
are outside of the purview of the goals of wepy and are not guaranteed
to be included in wepy indefinitely.


Please do not rely on anything in the util module to exist
indefinitely.

Any function is subject to deprecation at any release.

If you use these functions consider copying the source code.

If any of these functions or classes become reified and officially
part of wepy they will go in a sub-module dedicated to their purposes.

But these will likely (and should) be moved to other libraries as they
are more general purpose than wepy.

"""
