"""Resampler framework and library.

This sub-package provides an interface for implementing new resamplers
that are able to reuse existing distance metrics and decision classes.

Resamplers minimally must implement a single method 'resample'.

"""
